import json
from typing import Dict
from unittest import TestCase, main
from unittest.mock import MagicMock, patch
from gptbatcher import GPTBatcher, Question, Choice, Participant

import datetime
from unittest.mock import patch

from openai.types.chat import ChatCompletionMessage, ChatCompletionMessageToolCall
from openai.types.chat.chat_completion_message_tool_call import Function
from openai.types.chat.chat_completion import ChatCompletion, Choice as ChatCompletionChoice


def create_chat_completion(arguments: Dict[str, bool]) -> ChatCompletion:
    return ChatCompletion(
        id="foo",
        model="gpt-3.5-turbo",
        object="chat.completion",
        choices=[
            ChatCompletionChoice(
                finish_reason="stop",
                index=0,
                message=ChatCompletionMessage(
                    role="assistant",
                    tool_calls=[
                        ChatCompletionMessageToolCall(
                            id="foo",
                            type="function",
                            function=Function(
                                name="vote",
                                arguments=json.dumps(arguments),
                            )
                        )
                    ]
                ),
            )
        ],
        created=int(datetime.datetime.now().timestamp()),
    )

class TestGPTBatcher(TestCase):
    def setUp(self):
        self.api_key = 'test_key'
        self.rpm = 3
        self.model = 'gpt-3.5-turbo'
        self.temperature = 0.5

    @patch("openai.resources.chat.Completions.create")
    def test_ask(self, mock_create):
        mock_create.return_value = create_chat_completion({"Berlin": True, "Munich": False, "Hamburg": False, "Frankfurt": False})
        gpt_batcher = GPTBatcher(api_key=self.api_key, rpm=self.rpm, model=self.model, temperature=self.temperature)
        question = Question(
            label="Capital of Germany",
            prompt="What is the capital of Germany?",
            choices=[
                Choice(label="Berlin", prompt="Berlin"),
                Choice(label="Munich", prompt="Munich"),
                Choice(label="Hamburg", prompt="Hamburg"),
                Choice(label="Frankfurt", prompt="Frankfurt")
            ]
        )
        participant = Participant(label="Segment A", prompt="You are a 26 year old male from Birmingham, UK", weight=10)
        result = gpt_batcher.ask(question, [participant], 1)
        self.assertEqual(result.at[participant.label, "Berlin"], 1)

    @patch("openai.resources.chat.Completions.create")
    def test_failed_ask(self, mock_create):
        mock_create.return_value = create_chat_completion({"Berlin": False, "Munich": False, "Hamburg": False, "Frankfurt": False})
        gpt_batcher = GPTBatcher(api_key=self.api_key, rpm=self.rpm, model=self.model, temperature=self.temperature)
        question = Question(
            label="Capital of Germany",
            prompt="What is the capital of Germany?",
            choices=[
                Choice(label="Berlin", prompt="Berlin"),
                Choice(label="Munich", prompt="Munich"),
                Choice(label="Hamburg", prompt="Hamburg"),
                Choice(label="Frankfurt", prompt="Frankfurt")
            ]
        )
        participant = Participant(label="Segment A", prompt="You are a 26 year old male from Birmingham, UK", weight=10)
        with self.assertRaises(Exception):
            gpt_batcher.ask(question, [participant], 1)

    @patch("openai.resources.chat.Completions.create")
    def test_many_participants(self, mock_create):
        mock_create.return_value = create_chat_completion({"Berlin": True, "Munich": False, "Hamburg": False, "Frankfurt": False})
        gpt_batcher = GPTBatcher(api_key=self.api_key, rpm=self.rpm, model=self.model, temperature=self.temperature)
        question = Question(
            label="Capital of Germany",
            prompt="What is the capital of Germany?",
            choices=[
                Choice(label="Berlin", prompt="Berlin"),
                Choice(label="Munich", prompt="Munich"),
                Choice(label="Hamburg", prompt="Hamburg"),
                Choice(label="Frankfurt", prompt="Frankfurt")
            ]
        )
        participants = [
            Participant(label="Segment A", prompt="You are a 26 year old male from Birmingham, UK", weight=10),
            Participant(label="Segment B", prompt="You are a 45 year old female from London, UK", weight=10),
            Participant(label="Segment C", prompt="You are a 33 year old from Manchester, UK", weight=10),
        ]
        result = gpt_batcher.ask(question, participants, 3)
        self.assertEqual(result.sum()["Berlin"], 3)


if __name__ == '__main__':
    main()



import json
from openai import OpenAI
from pandas import DataFrame
from typing import List, Tuple
from gptbatcher.JobQueue import JobQueue
from gptbatcher.Sampler import Sampler
from gptbatcher.Question import Question
from gptbatcher.Choice import Choice
from gptbatcher.Participant import Participant

ParticipantChoice = Tuple[Participant, Choice]


async def ask_once(
    openai: OpenAI,
    model: str,
    temperature: float,
    question: Question,
    participant: Participant,
    function_name: str = "vote",
    function_description: str = "Vote on a poll",
) -> ParticipantChoice:
    completion = openai.chat.completions.create(
        model=model,
        temperature=temperature,
        messages=[
            {"role": "system", "content": participant.prompt},
            {"role": "user", "content": question.prompt},
        ],
        tools=[
            {
                "type": "function",
                "function": {
                    "name": function_name,
                    "description": function_description,
                    "parameters": {
                        "type": "object",
                        "properties": {
                            **{
                                choice.label: {
                                    "type": "boolean",
                                    "description": choice.prompt,
                                }
                                for choice in question.choices
                            }
                        },
                        "required": [choice.label for choice in question.choices],
                    },
                },
            }
        ],
        tool_choice={"type": "function", "function": {"name": function_name}},
    )

    choice = completion.choices[0]
    calls = choice.message.tool_calls
    if not calls or len(calls) != 1:
        raise Exception("Expected one tool call")

    json_string = calls[0].function.arguments
    results = json.loads(json_string)

    chosen = None
    for choice in question.choices:
        if choice.label in results and results[choice.label]:
            if chosen:
                raise Exception("Multiple choices selected")
            chosen = choice

    if not chosen:
        raise Exception("No choice selected")

    return (participant, chosen)


async def retry_ask_once(
    openai: OpenAI,
    model: str,
    temperature: float,
    question: Question,
    participant: Participant,
    function_name: str = "vote",
    function_description: str = "Vote on a poll",
    n: int = 3,
) -> ParticipantChoice:
    errors = []
    for _ in range(n):
        try:
            return await ask_once(openai, model, temperature, question, participant, function_name, function_description)
        except Exception as e:
            errors.append(e)
    raise Exception(", ".join(str(e) for e in errors))


class GPTBatcher:
    def __init__(
        self,
        api_key: str,
        rpm: int = 3,
        model: str = "gpt-3.5-turbo",
        temperature: float = 0.5,
        function_name: str = "vote",
        function_description: str = "Vote on a poll",
        retry: int = 3,
    ):
        if not api_key:
            raise Exception("API key required. You can find your API key at https://platform.openai.com/account/api-keys")
        self.openai = OpenAI(api_key=api_key)
        self.rpm = rpm
        self.model = model
        self.temperature = temperature
        self.function_name = function_name
        self.function_description = function_description
        self.retry = retry

    def ask(
        self, question: Question, participants: List[Participant], samples: int
    ) -> DataFrame:
        jobs = []
        sampler = Sampler(participants, lambda p: p.weight)
        for _ in range(samples):
            participant = sampler.sample()
            jobs.append(
                (
                    retry_ask_once,
                    (
                        self.openai,
                        self.model,
                        self.temperature,
                        question,
                        participant,
                        self.function_name,
                        self.function_description,
                        self.retry
                    ),
                )
            )
        job_queue = JobQueue(tokens=self.rpm, jobs=jobs)
        results: List[ParticipantChoice] = job_queue.run()
        if job_queue.errors:
            raise Exception(", ".join(str(e) for e in job_queue.errors))
        df = DataFrame(
            0,
            columns=[choice.label for choice in question.choices],
            index=[participant.label for participant in participants],
        )
        for participant, choice in results:
            df.at[participant.label, choice.label] += 1
        return df


import asyncio
import time
from typing import Callable, List, Any, Tuple


class JobQueue:
    def __init__(
        self, tokens: int, jobs: List[Tuple[Callable[..., Any], Tuple[Any, ...]]]
    ):
        self.tokens = asyncio.Semaphore(tokens)
        self.results = []
        self.errors = []
        self.jobs = jobs

    async def acquire_token(self):
        await self.tokens.acquire()

    def release_token(self):
        self.tokens.release()

    async def run_job(self, job: Callable[..., Any], *args) -> None:
        await self.acquire_token()
        start_time = time.time()
        try:
            result = await job(*args)
            self.results.append(result)
        except Exception as e:
            self.errors.append(e)
        finally:
            while len(self.results) + len(self.errors) < len(self.jobs):
                elapsed_time = time.time() - start_time
                if elapsed_time < 60:
                    await asyncio.sleep(1)
                else:
                    break
            self.release_token()

    async def run_all(self) -> List[Any]:
        tasks = [self.run_job(job, *args) for job, args in self.jobs]
        await asyncio.gather(*tasks)
        return self.results

    def run(self) -> List[Any]:
        return asyncio.run(self.run_all())



class Participant:
    def __init__(self, label: str, prompt: str, weight: float):
        self.label = label
        self.prompt = prompt
        self.weight = weight

    def __str__(self):
        return self.label



from typing import Any, Callable, List, TypeVar, Union
import random


class Sampler:
    def __init__(
        self, items: List[Any], weight_function: Callable[[Any], Union[int, float]]
    ):
        self.items = items
        self.weight_function = weight_function

    def sample(self) -> Any:
        total_weight = sum(self.weight_function(item) for item in self.items)
        random_weight_point = random.uniform(0, total_weight)
        current_weight = 0
        for item in self.items:
            current_weight += self.weight_function(item)
            if current_weight > random_weight_point:
                return item



from typing import List
from gptbatcher.Choice import Choice


class Question:
    def __init__(self, label: str, prompt: str, choices: List[Choice]):
        self.label = label
        self.prompt = prompt
        self.choices = choices

    def __str__(self):
        return self.label



from setuptools import setup, find_packages

setup(
    name='gptbatcher',
    version='0.2.0',
    url='https://github.com/CBM-Digital/GPTBatcher',
    author="James O'Toole",
    author_email='jamesotoole@cbmdigital.co.uk',
    description='Conduct surveys using GPT',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    packages=find_packages(),    
    install_requires=[
        'openai>=1.12.0',
        'pandas',
    ],
)
